//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/


#include "defs.h"
#include "spatial_pde.h"

int STEP_IDX=-1;

void execute(
	int DIM = 1,			// dimension
	int Q = 3,			// number of discrete velocities
	int model = SRT,		// collision model
	bool constvel = true,		// velocity field is constant in time and space (for ADE)
	int pde_strategy=NSE,		// Navier-Stokes (NSE) or Advection Diffusion (ADE) eqs
	bool use_symbolic_matrices=!true, // use symbolic matrice in the LBMAN algorithm
	int verbosity = 1,
	bool EFDEprint = true,		// export EFDE
	bool EPDEprint = true,		// export EPDE
	bool factor_coefs = false,	// use GiNaC::factor() when exporting coefficients to latex
	int ORDER = 4			// Taylor expansion order
)
{
	int step=1;
	SPATIAL_PDE core(DIM,Q);
	core.collision_model=model;
	core.verbosity=verbosity;
	core.constvel=constvel;
	core.EFDEprint=EFDEprint;
	core.factor_coefs=factor_coefs;
	core.ORDER = ORDER;
	core.pde_strategy = pde_strategy;
	core.use_symbolic_matrices = use_symbolic_matrices;

	core.allocate();
	sprintf(core.target_dir,"output_%s_d%dq%d%s_%s",(pde_strategy==NSE)?"nse":"ade",DIM,Q,(constvel)?"c":"n",core.model_desc);

	core.log("LBMAN: d=%d q=%d model=%d constvel=%s Taylor_order=%d pid=%d\n", DIM, Q, model, (constvel)?"yes":"no", ORDER, getpid());
	if (verbosity) core.log("|| step %02d: init & allocation\n",step++);
	mkdir(core.target_dir,0777);

	if (pde_strategy==ADE) core.use_NSE_strategies=false;
	core.use_NSE_strategies=true; 	// disable for D3Q27 for instance
//	core.sym_cs=(ex)1/sqrt((ex)3);	// lattice speed of sound
	core.use_NSE_omegas=true;


	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 01: compute LBMAN matrices
	//
	////////////////////////////////////////////////////////////////////////////////////////////

	if (core.use_NSE_strategies || core.collision_model != SRT)
	{
		if (STEP_IDX==0 || STEP_IDX==1)
		{
			if (verbosity) core.log("|| step %02d LBMAN symbolic matrices parallel\n", step++);
			if (!core.mark_exists("%s/_mark_compute_LBMAN_matrices_symbolic_save__finished",core.target_dir))
			{
				core.compute_LBMAN_matrices_symbolic_save_parallel();
				if (core.compute_LBMAN_matrices_symbolic_save_check());
					core.create_mark("%s/_mark_compute_LBMAN_matrices_symbolic_save__finished",core.target_dir);;
			}
		}
		if (STEP_IDX==0) while (!core.mark_exists("%s/_mark_compute_LBMAN_matrices_symbolic_save__finished",core.target_dir)) sleep(5);

		// once saved, subs!
		if (core.use_symbolic_matrices)
		{
			if (STEP_IDX==0 || STEP_IDX==1)
			{
				if (verbosity) core.log("|| step %02d LBMAN symbolic matrices simplify parallel  \n", step++);
				// load and mass simplify
				if (!core.mark_exists("%s/_mark_compute_LBMAN_matrices_symbolic_load_simplify_and_save__finished",core.target_dir))
				{
					core.compute_LBMAN_matrices_symbolic_load_simplify_and_save();
					// check whether all subs files are created
					if (core.compute_LBMAN_matrices_symbolic_load_simplify_and_save_check())
						core.create_mark("%s/_mark_compute_LBMAN_matrices_symbolic_load_simplify_and_save__finished",core.target_dir);
				}
			}
			if (STEP_IDX==0) while (!core.mark_exists("%s/_mark_compute_LBMAN_matrices_symbolic_load_simplify_and_save__finished",core.target_dir)) sleep(5);
		}
	} // is SRT model
	else
	{
		if (STEP_IDX==0 || STEP_IDX==1)
		{
			if (verbosity) core.log("|| step %02d LBMAN SRT matrices parallel    \n", step++);
			if (!core.mark("%s/_mark_compute_LBMAN_matrices_SRT_save",core.target_dir))
			{
				core.compute_LBMAN_matrices_SRT_save();
				core.create_mark("%s/_mark_compute_LBMAN_matrices_SRT_save__finished",core.target_dir);
			}
		}
		if (STEP_IDX==0) while (!core.mark_exists("%s/_mark_compute_LBMAN_matrices_SRT_save__finished",core.target_dir)) sleep(5);
	}

	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 02: compute EFDE coefficients
	//
	////////////////////////////////////////////////////////////////////////////////////////////


	if (STEP_IDX==0 || STEP_IDX==2)
	{
		if (verbosity) core.log("|| step %02d compute EFDE coefs parts parallel\n", step++);
		if (!core.mark_exists("%s/_mark_compute_EFDE_parts_save_parallel__finished",core.target_dir))
		{
			core.compute_EFDE_parts_save_parallel();
			if (core.compute_EFDE_parts_save_parallel_check())
				core.create_mark("%s/_mark_compute_EFDE_parts_save_parallel__finished",core.target_dir);;
		}
	}
	if (STEP_IDX==0) while (!core.mark_exists("%s/_mark_compute_EFDE_parts_save_parallel__finished",core.target_dir)) sleep(5);

	if (STEP_IDX==0 || STEP_IDX==2)
	{
		if (verbosity) core.log("|| step %02d combine & simplify EFDE coefs parts parallel\n", step++);
		if (!core.mark_exists("%s/_mark_compute_EFDE_coefs_save_parallel__finished",core.target_dir))
		{
			core.compute_EFDE_coefs_save_parallel();
			if (core.compute_EFDE_coefs_save_parallel_check())
				core.create_mark("%s/_mark_compute_EFDE_coefs_save_parallel__finished",core.target_dir);;
		}
	}
	if (STEP_IDX==0) while (!core.mark_exists("%s/_mark_compute_EFDE_coefs_save_parallel__finished",core.target_dir)) sleep(5);

/*
	// plot hollow-matrix-maker strategies for external investigation (MAXIMA/MAPLE format)
	if (core.use_NSE_strategies)
	{
		if (0)
		{
			for (int q=0;q<Q;q++)
			{
				cout << "output:"<<endl<<endl;
				core.find_optimal_pars(q, 0, Q);
				cout << "output:"<<endl<<endl;
			}
		}

		if (0)
		{
			for (int q=0;q<Q;q++)
			{
				core.find_optimal_pars(q, Q-2, Q);
			}
		}
		//return;
	}
*/

	if (STEP_IDX==0 || STEP_IDX==2)
	{
		if (EFDEprint)
		{
			if (verbosity) core.log("|| step %02d EFDE coefficients: print parallel\n", step++);
			for (int d=0;d<Q;d++)
			if (!core.mark("%s/_mark_export_EFDE_coefs_%d",core.target_dir,d))
			{
				core.create_mark("%s/_mark_export_EFDE_coefs_%d",core.target_dir,d);
				core.export_EFDE_coefs(d);
			}
		}
	}
	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 03: compute EPDE coefficients
	//
	////////////////////////////////////////////////////////////////////////////////////////////

	if (STEP_IDX==0 || STEP_IDX==3)
	{
		if (verbosity) core.log("|| step %02d EPDE coefficients: compute and save\n", step++);
		if (!core.mark_exists("%s/_mark_compute_EPDE_coefs_and_save__finished",core.target_dir))
		{
			core.compute_EPDE_coefs_and_save();
			if (core.compute_EPDE_coefs_and_save_check())
				core.create_mark("%s/_mark_compute_EPDE_coefs_and_save__finished",core.target_dir);
		}
	}

	if (STEP_IDX==0)
	while (!core.mark_exists("%s/_mark_compute_EPDE_coefs_and_save__finished",core.target_dir)) sleep(5);

	if (STEP_IDX==0 || STEP_IDX==7)
	{
		if (EPDEprint)
		{
			if (verbosity) core.log("|| step %02d EPDE coefficients: print parallel\n", step++);
			for (int d=0;d<Q;d++)
			if (!core.mark("%s/_mark_export_EPDE_coefs_%d",core.target_dir,d))
				core.export_EPDE_coefs(d);
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 04: spatial EPDE
	//
	////////////////////////////////////////////////////////////////////////////////////////////
	if (STEP_IDX==0 || STEP_IDX==4)
	{
		if (verbosity) core.log("|| step %02d spatial EPDE\n", step++);
		if (!core.mark("%s/_mark_spatial_EPDE", core.target_dir))
		{
			core.log("||| spatial EPDE start ");
			core.spatial_EPDE();
			core.create_mark("%s/_mark_spatial_EPDE__finished",core.target_dir);
		}
	}
	core.log("|| finished.");
}


int main(int argc, char* argv[])
{
	bool d1q3=!true;
	bool d2q5=!true;
	bool d2q9=!true;
	bool d3q7=!true;
	STEP_IDX=0;
	if (argc>=2) STEP_IDX=atoi(argv[1]);

	int cv=0;

	execute(1,3,CLBM1,false,ADE);

	if (d1q3)
	for (int m=0;m<=1;m++)
//	for (int cv=0;cv<=1;cv++)
	if (cv==0 || m==0)
	{
		int mod=(m==0)?ADE:NSE;
		execute(1,3,SRT,cv,mod);
		execute(1,3,MRT1,cv,mod);
		execute(1,3,CLBM1,cv,mod);
	}
	if (d2q5)
	{
		int mod=ADE;
		execute(2,5,SRT,cv,mod);
		execute(2,5,MRT1,cv,mod);
		execute(2,5,CLBM1,cv,mod);
	}
	if (d3q7)
	{
		int mod=ADE;
		execute(3,7,SRT,cv,mod);
		execute(3,7,MRT1,cv,mod);
		execute(3,7,CLBM1,cv,mod);
	}
	if (d2q9)
	{
		int mod=NSE;
		execute(2,9,SRT,cv,mod);
		execute(2,9,MRT1,cv,mod);
		execute(2,9,CLBM1,cv,mod);
	}
	cout << "success pid="<<getpid()<<endl;
	return 0;
}

